<?php

namespace Doctrine\DBAL\Platforms\Keywords;

/**
 * PostgreSQL 9.4 reserved keywords list.
 *
 * @deprecated Use {@see PostgreSQLKeywords} instead.
 */
class PostgreSQL94Keywords extends PostgreSQLKeywords
{
}
