<?php

namespace Doctrine\DBAL\Platforms;

/**
 * Provides the behavior, features and SQL dialect of the Microsoft SQL Server database platform
 * of the oldest supported version.
 *
 * @deprecated Use {@see SQLServerPlatform} instead.
 */
class SQLServer2012Platform extends SQLServerPlatform
{
}
